package matrix

import "fmt"
import "math/rand"

import "math"
import "../measuretime"

const Dim = 1200 //1e3

type matrix [Dim][Dim]float64

type rowResult struct {
	row      []float64
	rowIndex int
}

func MatrixMain() {
	fmt.Println("Dimensions: ", Dim)

	matA := new(matrix)
	matB := new(matrix)

	for i := 0; i < Dim; i++ {
		for j := 0; j < Dim; j++ {
			matA[i][j] = rand.Float64()
			matB[i][j] = rand.Float64()
		}
	}

	fmt.Println("Created factor matrices.")

	var seqTimer measuretime.Timer
	seqTimer.Pre()
	productSeq := matrixMultiplySeq(matA, matB)
	seqTimer.Post()
	seqTimer.PrintTimeConsumption("sequential algorithm")

	var seqTransTimer measuretime.Timer
	seqTransTimer.Pre()
	productSeqTrans := matrixMultiplySeqTrans(matA, matB)
	seqTransTimer.Post()
	seqTransTimer.PrintTimeConsumption("sequential transposed algorithm")

	var parTimer measuretime.Timer
	parTimer.Pre()
	productPar := matrixMultiplyPar(matA, matB)
	parTimer.Post()
	parTimer.PrintTimeConsumption("parallel algorithm")

	var parTransTimer measuretime.Timer
	parTransTimer.Pre()
	productParTrans := matrixMultiplyParTrans(matA, matB)
	parTransTimer.Post()
	parTransTimer.PrintTimeConsumption("parallel transposed algorithm")

	fmt.Println("computed product matrices, comparing results...")

	residuum := 0.0
	for i := 0; i < Dim; i++ {
		for j := 0; j < Dim; j++ {
			residuum += math.Abs(productSeq[i][j] - productPar[i][j])
		}
	}
	for i := 0; i < Dim; i++ {
		for j := 0; j < Dim; j++ {
			residuum += math.Abs(productSeq[i][j] - productSeqTrans[i][j])
		}
	}
	for i := 0; i < Dim; i++ {
		for j := 0; j < Dim; j++ {
			residuum += math.Abs(productSeq[i][j] - productParTrans[i][j])
		}
	}

	fmt.Println("Total residuum: ", residuum)
}

func matrixMultiplySeq(matA, matBUntransposed *matrix) [][]float64 {
	matB := matBUntransposed
	var product = make([][]float64, len(matA))
	for i := range product {
		product[i] = make([]float64, len(matB[0]))
		for j := 0; j < len(matB); j++ {
			var productItem float64
			for k := 0; k < len(matB); k++ {
				productItem += matA[i][k] * matB[k][j]
			}
			product[i][j] = productItem
		}
	}
	return product
}

func matrixMultiplySeqTrans(matA, matBUntransposed *matrix) [][]float64 {
	matB := transposeMatrix(matBUntransposed)
	var product = make([][]float64, len(matA))
	for i := range product {
		product[i] = make([]float64, len(matB[0]))
		for j := 0; j < len(matB); j++ {
			var productItem float64
			for k := 0; k < len(matB); k++ {
				productItem += matA[i][k] * matB[j][k]
			}
			product[i][j] = productItem
		}
	}
	return product
}

func matrixMultiplyPar(matA, matBUntransposed *matrix) [][]float64 {
	matB := matBUntransposed
	var product = make([][]float64, len(matA))
	rowChannel := make(chan rowResult)
	for i := range product {
		go computeRow(matA, matB, i, rowChannel)
	}

	for range product {
		result := <-rowChannel
		product[result.rowIndex] = result.row
	}
	close(rowChannel)

	return product
}

func computeRow(matA, matB *matrix, rowIndex int, rowChan chan rowResult) {
	var productRow = make([]float64, len(matA))
	for j := range productRow {
		var productItem float64
		for k := 0; k < len(matB); k++ {
			productItem += matA[rowIndex][k] * matB[k][j]
		}
		productRow[j] = productItem
	}

	var result rowResult
	result.row = productRow
	result.rowIndex = rowIndex
	rowChan <- result
}


func matrixMultiplyParTrans(matA, matBUntransposed *matrix) [][]float64 {
	matB := transposeMatrix(matBUntransposed)
	var product = make([][]float64, len(matA))
	rowChannel := make(chan rowResult)
	for i := range product {
		go computeRowTrans(matA, matB, i, rowChannel)
	}

	for range product {
		result := <-rowChannel
		product[result.rowIndex] = result.row
	}
	close(rowChannel)

	return product
}

func computeRowTrans(matA, matB *matrix, rowIndex int, rowChan chan rowResult) {
	var productRow = make([]float64, len(matA))
	for j := range productRow {
		var productItem float64
		for k := 0; k < len(matB); k++ {
			productItem += matA[rowIndex][k] * matB[j][k]
		}
		productRow[j] = productItem
	}

	var result rowResult
	result.row = productRow
	result.rowIndex = rowIndex
	rowChan <- result
}

func transposeMatrix(matA *matrix) *matrix {
	mat := new(matrix)
	for i := 0; i < Dim; i++ {
		for j := 0; j < Dim; j++ {
			mat[i][j] = matA[j][i]
		}
	}
	return mat
}
