import java.io.IOException;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.BindException;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue; 

public class ChannelFactory extends Thread {
    public final static int DEFAULT_PORT = 4711;
    private final int port;
    private final BlockingQueue<ObjectChannel> buf;
    private volatile ServerSocket srv;
    private volatile boolean srvrun = false;
    private volatile boolean srvstart = false;

    public ChannelFactory() {
        this(DEFAULT_PORT);
    }

    public ChannelFactory(int p) {
        buf = new LinkedBlockingQueue<ObjectChannel>(/*infinite*/);
        if (p <= 0) {
            port = DEFAULT_PORT;
        } else {
            port = p;
        }
        try {
            srv = new ServerSocket(port);
            //this.start(); moved to init and getChannel
            System.out.println("server bound to port " + port);
        } catch (BindException e) {
            srv = null;
            System.out.println("server not started, port used " + port);
        } catch (IOException e) {
            System.out.println("IOException " + e);
        }
    }

    public String toString() {
        return "" + this.getClass().getSimpleName() + "(" + srv + ", buf = " + buf.size() + ")";
    }

    public void init() {
        if (srv != null && ! srvstart  ) {
            this.start();
            srvstart = true;
            System.out.println("ChannelFactory at " + srv);
        }
    }

    public ObjectChannel getChannel() throws InterruptedException {
        if (srv == null) {
            if (srvrun) {
                throw new IllegalArgumentException("dont call when no server listens");
            }
        } else if ( ! srvstart  ) {
            init();
        }
        return buf.take();
    }

    public ObjectChannel getChannel(String h, int p) throws IOException {
        if (p <= 0) {
            p = port;
        }
        ObjectChannel c = null;
        int i = 0;
        int delay = 5; // 50
        System.out.println("connecting to " + h);
        while (c == null) {
            try {
                c = new ObjectChannel(new Socket(h, p));
            } catch (IOException e) {
                i++;
                if (i % 50 == 0) {
                    delay += delay;
                    System.out.println("Server on " + h + " not ready in " + delay + "ms");
                }
                System.out.println("Server on " + h + " not ready in " + delay + "ms");
                try {
                    Thread.sleep(delay);
                } catch (InterruptedException w) {
                    Thread.currentThread().interrupt();
                    throw new IOException("Interrupted during IO wait " + w);
                }
            }
        }
        System.out.println("connected, iter = " + i);
        return c;
    }

    @Override
    public void run() {
        if (srv == null) {
            return; // nothing to do
        }
        srvrun = true;
        while (true) {
            try {
                System.out.println("waiting for connection");
                Socket s = srv.accept();
                if (this.isInterrupted()) {
                    srvrun = false;
                    return;
                }
                System.out.println("connection accepted");
                ObjectChannel c = new ObjectChannel(s);
                buf.put(c);
            } catch (IOException e) {
                srvrun = false;
                return;
            } catch (InterruptedException e) {
                srvrun = false;
                return;
            }
        }
    }

    public void terminate() {
        if ( ! srvstart ) {
            System.out.println("server not started");
            return; 
        }
        this.interrupt();
        try {
            if (srv != null) {
                srv.close();
                srvrun = false;
            }
            this.interrupt();
            while (!buf.isEmpty()) {
                System.out.println("closing unused SocketChannel");
                ObjectChannel c  = buf.poll();
                if ( c != null ) {
                    c.close();
                }
            }
        } catch (IOException e) {
        }
        try {
            this.join();
        } catch (InterruptedException e) {
        }
        System.out.println("ChannelFactory terminated");
    }
}

