
import java.util.Set;

import java.io.IOException;
import java.net.BindException;
import java.net.InetSocketAddress;
import java.net.Socket;
import java.net.ServerSocket;

import java.nio.channels.Selector;
import java.nio.channels.SelectionKey;
import java.nio.channels.ServerSocketChannel;
//import java.nio.channels.SocketChannel;


/**
 * ChannelFactoryNio.
 * Socket channel factory using java.nio and no Thread.
 *
 * Usage: ChannelFactoryNio(port).
 * @author Heinz Kredel.
 */
public class ChannelFactoryNio extends Thread {
  
/**
 * default port of socket. 
 */
  public final static int DEFAULT_PORT = 4711;

/**
 * wait time 500ms. 
 */
  public final static int WAIT_TIME = 500;

/**
 * port of socket. 
 */
  private final int port;

/**
 * local selector.
 */
  private Selector selector = null;

/**
 * local server key.
 */
  private SelectionKey serverkey = null;

/**
 * local server socket channel.
 */
  private ServerSocketChannel server = null;


/**
 * Constructs a ChannelFactory using Nio. 
 * @param p port.
 */
  public ChannelFactoryNio(int p) {
    if ( p <= 0 ) { 
       port = DEFAULT_PORT; 
    } else { 
       port = p; 
    }
    try {
        selector = Selector.open();

        server = ServerSocketChannel.open();
        server.configureBlocking(false);
        serverkey = server.register(selector,SelectionKey.OP_ACCEPT);
        server.socket().bind( new InetSocketAddress(port) );

        System.out.println("server started on port "+port);
    } catch (BindException e) { 
        System.out.println(e);
        serverkey.cancel();
        //server.close();
    } catch (IOException e) { 
        e.printStackTrace(); 
    }
  }


/**
 * Constructs a ChannelFactory Nio.
 * on the DEFAULT_PORT.
 */
  public ChannelFactoryNio() {
    this(DEFAULT_PORT); 
  }


/**
 * Get a new socket channel from a server socket.
 */
  public SocketChannel getChannel() 
                       throws IOException {
    SocketChannel c = null;
    java.nio.channels.SocketChannel s;
    while (c == null) {
        try {
            System.out.println("waiting for connection");
            selector.select(); // blocking

            Set<SelectionKey> keys = selector.selectedKeys();
            for ( SelectionKey k: keys ) {
                if ( k == serverkey ) {
                   if ( k.isAcceptable() ) { // just in case
                      keys.remove(k);
                      s = server.accept();
                      s.configureBlocking(true);
                      c = new SocketChannel( s.socket() );
                      System.out.println("connection accepted");
                      break; // ignore other keys
                   }
                } else { // ignore client keys
                    selector.wakeup();
                }
            }
        } catch (IOException e) {
          System.out.println(e);
          try {   // wait for next client to connect
              Thread.currentThread().sleep( WAIT_TIME );
          } catch (InterruptedException ignored) { } 
        }
    }
    return c;
  }


/**
 * Get a new socket channel to a given host.
 * @param h hostname.
 * @param p port.
 */
  public SocketChannel getChannel(String h, int p) 
                                  throws IOException {
    if ( p <= 0 ) { 
       p = port; 
    } 
    SelectionKey clientkey = null;
    SocketChannel c = null;
    java.nio.channels.SocketChannel client;
    System.out.println("connecting to "+h);
    while (c == null) {
        try {
            client = java.nio.channels.SocketChannel.open();
            client.configureBlocking(false);
            clientkey = client.register(selector,SelectionKey.OP_CONNECT);
            client.connect( new InetSocketAddress(h,p) );

            selector.select(); // blocking

            Set<SelectionKey> keys = selector.selectedKeys();
            for ( SelectionKey k: keys ) {
                if ( k == clientkey ) {
                   if ( k.isConnectable() ) { // if server ready
                       keys.remove(k);
                       k.cancel();     // required to make blockable
                       client = (java.nio.channels.SocketChannel) k.channel();
                       client.finishConnect();
                       client.configureBlocking(true);
                       c = new SocketChannel( client.socket() );
                       break; // ignore other keys
                   }
                } else { // ignore server keys
                    selector.wakeup();
                }
            }
        } catch (IOException e) {
          System.out.println(e);
          System.out.println("Server on "+h+" not ready");
          try { // wait server ready
              Thread.currentThread().sleep( WAIT_TIME );
          } catch (InterruptedException ignored) { } 
        }
    }
    System.out.println("connected");
    return c;
  }


/**
 * Close the Channel Factory.
 */
  public void close() {
    try {
        serverkey.cancel();
        selector.close();
        server.close();
    } catch (IOException e) { 
        e.printStackTrace();
    }
  }

}

