//package edu.unima.ky.parallel;

import java.io.IOException;
import java.nio.channels.SocketChannel;
import java.nio.ByteBuffer;

/**
 * ObjectBufferChannel.
 * This class creates a communication channel using a socket.
 * @author Heinz Kredel.
 */

public class ObjectBufferChannel implements ObjectChannel {

  /**
   * default capacity for ByteBuffer.
   */
  final static int DEFAULT_CAPACITY = 4*1024; // 4k

  /**
   * heap ByteBuffer.
   */
  final ByteBuffer hbb;

  /**
   * object ByteBuffer wraper.
   */
  final ObjectBuffer ob;

  /**
   * socket channel.
   */
  final SocketChannel soc;


  /**
   * Constructs a socket channel on the given socket s.
   * @param s socket channel.
   * @param cap capacity of ByteBuffer.
   */
  public ObjectBufferChannel(SocketChannel s, int cap) 
                             throws IOException {
    soc = s;
    hbb = ByteBuffer.allocate(cap);
    // hbb.clear(); // or hbb.rewind() ?
    ob = new ObjectBuffer( hbb );
  }

  /**
   * Constructs a socket channel on the given socket s.
   * @param s socket channel.
   */
  public ObjectBufferChannel(SocketChannel s) 
                             throws IOException {
      this(s,DEFAULT_CAPACITY);
  }

  /**
   * Sends an object.
   * @param v Object to send.
   */
  public void send(Object v) throws IOException {
    synchronized (hbb) {
      hbb.clear();
      ob.put(v);
      hbb.flip();
      soc.write( hbb );
    }
  }


  /**
   * Receives an object.
   * @return received Object.
   */
  public Object receive() throws IOException, ClassNotFoundException {
    Object v = null;
    synchronized (hbb) { // use different buffer
      hbb.clear(); // if remaining() == 0
      soc.read( hbb );
      hbb.flip();
      v = ob.get();
    }
    return v;
  }


  /**
   * Closes the channel.
   */
  public void close() {
    if (soc != null) {
      try { soc.close(); } catch (IOException e) { }
    }
  }

}
