
import java.util.ArrayList;

/**
 * A parallel algorithm for an euclidean 2d TSP Problem
 */

public class ParByteLocalTSP implements TSPInf {

    protected Graph graph;
    protected Point[] points;
    // protected Path bestPath; // beware of threads
    protected BestByteLocStore best;

    protected int number = 0;

    public ParByteLocalTSP(Point[] p, int th) {
        this.points = p;
        graph = new PlaneGraph(p);
        number = th;
    }

    public Path actualBest() {
        return best.getPath();
    }

    public void setBest(Path b) {
        if ( b == null ) {
           return;
        }
        if ( best == null ) {
           // best.setPath( b );
           return;
        }
        if ( b.cost() < best.getPath().cost() ) {
           best.setPath( b );
        }
    }


    public Path getBest() {
        return getBest(Long.MAX_VALUE);
    }

    public Path getBest(long max) {
        Path bestPath = PathByteArray.standardPath(graph); // standard path
        if ( bestPath.length() <= 2 ) {
            return bestPath;
        }
        best = new BestByteLocStore( bestPath, 0, number );
        Path path = PathByteArray.emptyPath(graph); // empty path
        Deque globalStack = new Deque(10000);
        try {
            globalStack.push( path );
        } catch (InterruptedException e) {
            e.printStackTrace();
        }
        Thread[] t = new Thread[number];
        for ( int i = 0; i < number; i++ ) {
            t[i] = new RunByteLocTSP(globalStack,best,max);
            t[i].start();
        }
        try {
            for ( int i = 0; i < number; i++ ) {
                t[i].join();
            }
        } catch (InterruptedException e) {
            e.printStackTrace();
        }
        return best.getPath();
    }

    public void getBest(Path path) {
        System.out.println("method getBest() in ParByteLocalTSP not implemented");
    }

}

class BestByteLocStore {

    private Path path;
    private int idlers;
    public final int threads;

    public BestByteLocStore(Path b, int i, int t) {
        path = b;
        idlers = i;
        threads = t;
    }

    public Path getPath() {
        return path;
    }

    public void setPath(Path p) {
        path = p;
    }

    public synchronized void incIdle() {
        idlers++;
    }

    public synchronized void decIdle() {
        idlers--;
    }

    public synchronized boolean allIdle() {
        return (idlers == threads);
    }
}


class RunByteLocTSP extends Thread {

    private BestByteLocStore best;
    private Deque globalStack;
    private long maxiter;
    private long iter;
    private long pushes;
    private ArrayList stack;
    public final int COUNT;

    public RunByteLocTSP(Deque s, BestByteLocStore b, long max) {
        globalStack = s;
        best = b;
        COUNT = b.getPath().maxsize()+5; //*2 / +5 ?
        maxiter = max;
        stack = new ArrayList(500);
    }


    public void run() {
        iter = 0;
        pushes = 0;
        boolean running = true;
        while (running) {
            Object o = null;
            if ( stack.isEmpty() ) {
               try {
                  best.incIdle();
                  if ( globalStack.empty() && best.allIdle() ) {
                      for ( int i = 0; i < best.threads-1; i++) {
                          globalStack.push(null);
                      }
                      best.decIdle();
                      break; //return;
                  }
                  o = globalStack.pop();
                  best.decIdle();
               } catch (InterruptedException e) {
                  e.printStackTrace();
                  running = false;
               }
            } else {
                o = stack.remove( stack.size()-1 );
            }
            if ( o == null ) {
                break; //return;
            }
            Path path = (Path)o;
            getBest( path );
        }
        System.out.println("iterations = " + iter + " pushes " + pushes);
    }


    public void getBest(Path path) {
        Path[] ps = path.nextPaths(); 
        iter += ps.length;
        double bestCost = best.getPath().cost();
        if ( path.length()+1 == path.maxsize() ) {
           for ( int i = 0; i < ps.length; i++ ) { // branch
               if ( ps[i].cost() < bestCost ) {
                 synchronized ( best ) {
                     if ( ps[i].cost() < best.getPath().cost() ) {
                        bestCost = ps[i].cost();
                        best.setPath( ps[i] );
                        System.out.println("par best = " + ps[i]);
                   }
                 }
               }
           }
           return;
        }
        if ( iter >= maxiter ) {
            return;
        }
        if ( path.length()+5 >= path.maxsize() ) {
           for ( int i = 0; i < ps.length; i++ ) { // branch
               if ( ps[i].cost() < bestCost ) { // cut
                   //System.out.print("#");
                   getBestRec( ps[i].copyMax(), path.length()+1 ); // changes best and bestPath
                  //stack.add( ps[i] );
               }
           }
           return;
        }
        for ( int i = 0; i < ps.length; i++ ) { // branch
            if ( ps[i].cost() < bestCost ) { // cut
               // getBest( ps[i] ); // changes best and bestPath
               //System.out.print("-");
               stack.add( ps[i] );
            }
        }
        if ( stack.size() > COUNT && path.length() <= 3 ) {
            if ( globalStack.empty() ) {
               Object o = stack.remove( 0 ); //stack.size()-1, depth first
               pushes++;
               try {
                   globalStack.push( o ); // beware of deadlock
               } catch (InterruptedException e) {
                   e.printStackTrace();
               }
            }
        }
        return;
    }

    public void getBestRec(Path path, int depth) {
        double bestCost = best.getPath().cost();
        for ( int k = 0; k < path.maxsize()-depth; k++ ) { // branch
            Path ps = path.nextPath(k); 
            iter++;
            if ( depth+1 == path.maxsize() ) {
               if ( ps.cost() < bestCost ) {
                  synchronized ( best ) {
                     if ( ps.cost() < best.getPath().cost() ) {
                        bestCost = ps.cost();
                        best.setPath( (Path)ps.clone() ); // had error
                        System.out.println("par best rec = " + ps);
                     }
                  }
               }
            }
            if ( iter >= maxiter ) {
               return;
            }
            if ( ps.cost() < bestCost ) { // cut
                getBestRec( ps, depth+1 ); // changes best and bestPath
                //stack.add( ps[i] );
            }
        }
        return;
    }


}
